// Used by vncconfig.rc

#include <rfb_win32/resource.h>

#define IDR_MANIFEST                    1
#define IDI_ICON                        101
#define IDD_DIALOG1                     102
#define IDD_DIALOG2                     103
#define IDD_CONNECTIONS                 105
#define IDD_HOOKING                     106
#define IDD_VNC_PASSWD                  107
#define IDD_AUTH_VNC_PASSWD             107
#define IDD_LEGACY                      108
#define IDD_CONN_HOST                   109
#define IDD_SHARING                     110
#define IDD_INPUTS                      111
#define IDR_TRAY                        112
#define IDD_ABOUT                       113
#define IDI_CONNECTED                   115
#define IDD_DESKTOP                     116
#define IDC_EDIT1                       1000
#define IDC_PORT                        1000
#define IDC_PASSWORD1                   1000
#define IDC_HOST_PATTERN                1000
#define IDC_AUTH_VNC_PASSWD             1009
#define IDC_USEHOOKS                    1011
#define IDC_POLLCONSOLES                1012
#define IDC_COMPAREFB                   1013
#define IDC_IDLE_TIMEOUT                1015
#define IDC_HOSTS                       1016
#define IDC_HOST_ADD                    1017
#define IDC_HOST_REMOVE                 1018
#define IDC_HOST_UP                     1019
#define IDC_BUTTON4                     1020
#define IDC_HOST_DOWN                   1020
#define IDC_AUTH_INPUTONLY_PASSWD       1020
#define IDC_HOST_EDIT                   1021
#define IDC_PASSWORD2                   1022
#define IDC_LEGACY_IMPORT               1023
#define IDC_ALLOW                       1024
#define IDC_DENY                        1025
#define IDC_SHARE_ALWAYS                1030
#define IDC_SHARE_NEVER                 1031
#define IDC_SHARE_CLIENT                1032
#define IDC_DISCONNECT_CLIENTS          1033
#define IDC_ACCEPT_KEYS                 1034
#define IDC_ACCEPT_PTR                  1035
#define IDC_ACCEPT_CUTTEXT              1036
#define IDC_SEND_CUTTEXT                1037
#define IDC_PROTOCOL_3_3                1038
#define IDC_DESCRIPTION                 1039
#define IDC_BUILDTIME                   1040
#define IDC_VERSION                     1041
#define IDC_COPYRIGHT                   1042
#define IDC_BL_THRESHOLD                1046
#define IDC_BL_TIMEOUT                  1047
#define IDC_AFFECT_SCREENSAVER          1048
#define IDC_LOCALHOST                   1049
#define IDC_DISABLE_LOCAL_INPUTS        1050
#define IDC_QUERY_CONNECT               1055
#define IDC_DISCONNECT_NONE             1056
#define IDC_DISCONNECT_LOCK             1057
#define IDC_DISCONNECT_LOGOFF           1058
#define IDC_REMOVE_WALLPAPER            1059
#define IDC_DISABLE_EFFECTS             1061
#define IDC_CAPTUREBLT                  1062
#define IDC_QUERY                       1064
#define IDC_USEPOLLING                  1066
#define IDC_QUERY_LOGGED_ON             1069
#define IDC_AUTH_ADMIN_PASSWD           1076
#define IDC_AUTH_VIEWONLY_PASSWD        1077
#define IDC_AUTH_ADMIN_ENABLE           1078
#define IDC_AUTH_VIEWONLY_ENABLE        1079
#define IDC_AUTH_INPUTONLY_ENABLE       1080
#define IDC_RFB_ENABLE                  1082
#define IDC_LOAD_CERT                   1087
#define IDC_LOAD_CERTKEY                1088
#define IDC_RAW_KEYBOARD                1089
#define ID_OPTIONS                      40001
#define ID_CLOSE                        40002
#define ID_ABOUT                        40003

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        117
#define _APS_NEXT_COMMAND_VALUE         40004
#define _APS_NEXT_CONTROL_VALUE         1083
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif
